package ch.unibe.piccola;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Properties;
import ch.unibe.piccola.code.Code;
import ch.unibe.piccola.compiler.DefaultCompiler;
import ch.unibe.piccola.compiler.Simplifier;
import ch.unibe.piccola.parser.Parser;
import ch.unibe.piccola.parser.ParserException;
import ch.unibe.piccola.parser.ParserFactory;
import ch.unibe.piccola.source.SourceCode;
import ch.unibe.piccola.source.StringSourceCode;
import ch.unibe.piccola.source.URLSourceCode;
import ch.unibe.piccola.syntax.Node;
import ch.unibe.piccola.syntax.Syntax;
import ch.unibe.piccola.type.TypeInference;
import ch.unibe.piccola.vm.AgentException;
import ch.unibe.piccola.vm.Form;
import ch.unibe.piccola.vm.VM;
import ch.unibe.piccola.vm.VMPool;

/**
 * The Piccola runtime host. This class represents a Piccola runtime
 * environment.
 *
 * @author   Stefan Kneubuehl
 * @version  3.4c (Dec 16, 2002)
**/
public class Host extends VMPool
{

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String LABEL_DYNAMIC = Syntax.labelDynamic();

    private static final String LABEL_HOMEURL = Syntax.labelHomeURL();

    private static final String LABEL_MAIN = Syntax.labelMain();

    private static final String LABEL_META = Syntax.labelMeta();

    private static final String PROPERTY_FILE = "properties";

    private static final String PRELUDE_FILE = "lib/prelude.picl";

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Creates a new Piccola host.
    **/
    public Host()
    {
        m_classLoader = getClass().getClassLoader();
        m_parser = ParserFactory.INSTANCE.newJavaCCParser();
        m_compiler = new DefaultCompiler();
        m_out = System.out;
    }

    /**
     * Initializes the Piccola environment. Loads properties from the Piccola
     * property file, and initializes the standard static and dynamic context
     * by executing the prelude script.
    **/
    public void initialize()
    {
        initProperties();
        initContext();
    }

    /**
     * Evaluates the specified Piccola script in this Piccola environment.
     *
     * @param  name the name of the Piccola script
     * @return the form value of the specified Piccola script
    **/
    public final Form evaluateScript(String name)
    {
        return serviceFromScript(name, m_standardContext).service().evaluate(m_standardDynamic);
    }

    /**
     * Executes the specified Piccola script in this Piccola environment.
     * Evaluate the script, then executes the <code>main</code> service, if
     * present.
     *
     * @param  name the name of the Piccola script
    **/
    public final void executeScript(String name)
    {
        Form script = evaluateScript(name);
        if (script.exists(LABEL_MAIN)) {
            script.project(LABEL_MAIN).service().run(m_standardDynamic);
        }
    }

    /**
     * Executes the specified Piccola source code in this Piccola environment.
     *
     * @param  code the Piccola source code.
    **/
    public final void executeCode(String code)
    {
        serviceFromString(code, m_standardContext).service().run(m_standardDynamic);
    }

    /**
     * Tries to find a resource specified by it's name and possibly a relative
     * path. Returns the resource's URL or null if the resource cannot be
     * found.
     *
     * @param  name the name of the resource to be found.
     * @return the URL of the resource or null, if the resource cannot be
     *         found.
    **/
    public final URL findResource(String name)
    {
        URL result = null;
        try {
            result = new URL(name);
        }
        catch (MalformedURLException ex) { }
        if (null == result) {
            result = m_classLoader.getResource(name);
        }
        if (null == result) {
            String pwd = System.getProperty("user.dir", ".");
            File file = new File(pwd + "/" + name);
            if (file.exists()) {
                try {
                    result = file.toURL();
                }
                catch (MalformedURLException ex) { }
            }
        }
        return result;
    }

    /**
     * Returns the value of a Piccola system property with the given key.
     * The Piccola system properties is a superset of the Java system
     * properties. They always include the following keys:
     * <ul>
     *  <li><code>piccola.prelude</code> Piccola prelude script</li>
     *  <li><code>piccola.prelude.url</code> Piccola prelude script URL</li>
     *  <li><code>piccola.version</code> Piccola Runtime Environment
     *      version</li>
     *  <li><code>piccola.script</code> Piccola default script</li>
     *  <li><code>java.version</code> Java Runtime Environment version</li>
     *  <li><code>java.vendor</code> Java Runtime Environment vendor</li>
     *  <li><code>java.vendor.url</code> Java vendor URL</li>
     *  <li><code>java.home</code> Java installation directory</li>
     *  <li><code>java.vm.specification.version</code> Java Virtual
     *      Machine specification version</li>
     *  <li><code>java.vm.specification.vendor</code> Java Virtual
     *      Machine specification vendor</li>
     *  <li><code>java.vm.specification.name</code> Java Virtual
     *      Machine specification name</li>
     *  <li><code>java.class.version</code>Java class format version
     *      number</li>
     *  <li><code>java.class.path</code> Java class path</li>
     *  <li><code>java.library.path</code> List of paths to search when loading
     *      libraries</li>
     *  <li><code>java.io.tmpdir</code> Default temp file path</li>
     *  <li><code>java.compiler</code> Name of JIT compiler in use</li>
     *  <li><code>java.ext.dirs</code> Path of extension directory or
     *      directories</li>
     *  <li><code>os.name</code> Operating system name</li>
     *  <li><code>os.arch</code> Operating system architecture</li>
     *  <li><code>os.version</code> Operating system version</li>
     *  <li><code>file.separator</code> File separator ("/" un UNIX)</li>
     *  <li><code>path.separator</code> Path separator (":" un UNIX)</li>
     *  <li><code>line.separator</code> Line separator ("\n" un UNIX)</li>
     *  <li><code>user.name</code> User's account name</li>
     *  <li><code>user.home</code> User's home directory</li>
     *  <li><code>user.dir</code> User's current working directory</li>
     * </ul>
     * @see java.lang.System#getProperties()
    **/
    public final String getProperty(String key)
    {
        return m_properties.getProperty(key);
    }

    // ------------------------------------------------------------------------
    // default
    // ------------------------------------------------------------------------

    Form serviceFromScript(String name, Form context)
    {
        URL url = findResource(name);
        if (null == url) {
            throw new AgentException("Script \"" + name + "\" not found.");
        }
        return serviceFromURL(url, context);
    }

    Form serviceFromString(String code, Form context)
    {
        return serviceFromSourceCode(new StringSourceCode(code), context);
    }

    Form serviceFromURL(URL url, Form context)
    {
        context = context.extend(up(url).bind(LABEL_HOMEURL));
        return serviceFromSourceCode(new URLSourceCode(url), context);
    }

    String typeFromString(String code)
    {
        return TypeInference.apply(Simplifier.apply(parse(new StringSourceCode(code)))).toString();
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private ClassLoader m_classLoader;

    private DefaultCompiler m_compiler;

    private PrintStream m_out;

    private Parser m_parser;

    private Properties m_properties;

    private Form m_standardContext;

    private Form m_standardDynamic;

    private Node parse(SourceCode sourceCode)
    {
        try {
            return m_parser.parse(sourceCode);
        }
        catch (IOException ex) {
            throw new AgentException(ex);
        }
        catch (ParserException ex) {
            throw new AgentException(ex);
        }
    }

    private Code compile(SourceCode sourceCode)
    {
        return m_compiler.compile(parse(sourceCode));
    }

    private Form serviceFromSourceCode(SourceCode sourceCode, Form staticContext)
    {
        return newService(compile(sourceCode), staticContext);
    }

    private void initContext()
    {
        m_standardContext = up(new RuntimeHost(this));
        m_standardContext = m_standardContext.extend(newMetaService().bind(LABEL_META));
        m_standardDynamic = m_factory.emptyForm();

        m_standardContext = evaluateScript(getProperty("piccola.prelude"));
        m_standardDynamic = m_standardContext.project(LABEL_DYNAMIC);
    }

    private final void initProperties()
    {
        Properties properties = new Properties();
        URL url = findResource(PROPERTY_FILE);
        if (null != url) {
            try {
                properties.load(url.openStream());
            }
            catch (IOException ex) { }
        }
        m_properties = new Properties(System.getProperties());
        m_properties.put("piccola.prelude", PRELUDE_FILE);
        m_properties.putAll(properties);
        m_properties.put("piccola.prelude.url", findResource(getProperty("piccola.prelude")).toExternalForm());
    }

}

