package ch.unibe.piccola.bridge;

/**
 * Implements wrapper methods for primitive number operations. These
 * operations are defined for all <code>Number</code> objects, but the
 * methods operate internally only with the following primitive number types:
 *
 * <ul>
 *   <li> double
 *   <li> float
 *   <li> long
 *   <li> int
 * </ul>
 *
 * Arguments of other types are converted to one of the listed types before
 * the operation. The standard conversions are the following:
 *
 * <ul>
 *   <li><code>Integer</code>, <code>Short</code>, and
 *       <code>Byte</code> objects are converted to <code>int</code>.
 *   <li><code>Long</code> objects are converted to <code>long</code>.
 *   <li><code>Float</code> objects are converted to <code>float</code>.
 *   <li><code>Double</code> and other <code>Number</code> objects are
 *       converted to <code>double</code>.
 * </ul>
 * If the result of the operation is a number, the object returned is an
 * instance of <code>Integer</code>, <code>Long</code>, <code>Float</code>,
 * or <code>Double</code>, depending on the larger internal operand type.
 * Exceptions to these rules are documented with the concerned methods.
 *
 * @author   Stefan Kneubuehl
 * @version  4.0 (Jun 09, 2003)
**/
public class NumberOp
{

    // ------------------------------------------------------------------------
    // public static
    // ------------------------------------------------------------------------

    /**
     * Returns the absolute value of the specified number.
     * values.
     *
     * @param  a operand
     * @return the absolute (positive) value of <codea</code>.
    **/
    public static Number abs(Number a)
    {
        switch (getType(a)) {
            case UNKNOWN:
            case DOUBLE: return new Double(Math.abs(a.doubleValue()));
            case FLOAT: return new Float(Math.abs(a.floatValue()));
            case LONG: return new Long(Math.abs(a.longValue()));
            case INTEGER: return new Integer(Math.abs(a.intValue()));
            default: throw new Error("Unexpected number type.");
        }
    }

    /**
     * Returns the quotient of the two arguments.
     *
     * @param  a dividend
     * @param  b divisor
     * @return the quotient of <code>a</code> and <code>b</code>.
    **/
    public static Number divide(Number a, Number b)
    {
        switch (returnType(a, b)) {
            case UNKNOWN:
            case DOUBLE: return new Double(a.doubleValue() / b.doubleValue());
            case FLOAT: return new Float(a.floatValue() / b.floatValue());
            case LONG: return new Long(a.longValue() / b.longValue());
            case INTEGER: return new Integer(a.intValue() / b.intValue());
            default: throw new Error("Unexpected number type.");
        }
    }

    /**
     * Compares a number with an object.
     *
     * @param  a the number to be compared with an object
     * @param  b the object to be compared with the number
     * @return <code>true</code>, if <code>b</code> is a number and
     *         the values of <code>a</code> and <code>b</code> are equal,
     *         <code>false</code> otherwise.
    **/
    public static boolean equals(Number a, Object object)
    {
        if (object instanceof Number) {
            Number b = (Number) object;
            switch (returnType(a, b)) {
                case UNKNOWN:
                case DOUBLE: return a.doubleValue() == b.doubleValue();
                case FLOAT: return a.floatValue() == b.floatValue();
                case LONG: return a.longValue() == b.longValue();
                case INTEGER: return a.intValue() == b.intValue();
                default: throw new Error("Unexpected number type.");
            }
        }
        else {
            return false;
        }
    }

    /**
     * Returns the difference between the two arguments
     *
     * @param  a the first operand of the subraction
     * @param  b the second operand
     * @return the difference between <code>a</code> and <code>b</code>.
    **/
    public static Number minus(Number a, Number b)
    {
        switch (returnType(a, b)) {
            case UNKNOWN:
            case DOUBLE: return new Double(a.doubleValue() - b.doubleValue());
            case FLOAT: return new Float(a.floatValue() - b.floatValue());
            case LONG: return new Long(a.longValue() - b.longValue());
            case INTEGER: return new Integer(a.intValue() - b.intValue());
            default: throw new Error("Unexpected number type.");
        }
    }

    /**
     * Returns the remainder of the second argument in the first argument.
     * Arguments of floating-point or unkown type are converted to longs
     * before the operation.
    **/
    public static Number mod(Number a, Number b)
    {
        switch (returnType(a, b)) {
            case UNKNOWN:
            case DOUBLE:
            case FLOAT:
            case LONG: return new Long(a.longValue() % b.longValue());
            case INTEGER: return new Integer(a.intValue() % b.intValue());
            default: throw new Error("Unexpected number type.");
        }
    }

    public static Number negate(Number a)
    {
        switch (getType(a)) {
            case UNKNOWN:
            case DOUBLE: return new Double(-a.doubleValue());
            case FLOAT: return new Float(-a.floatValue());
            case LONG: return new Long(-a.longValue());
            case INTEGER: return new Integer(-a.intValue());
            default: throw new Error("Unexpected number type.");
        }
    }

    public static Number plus(Number a, Number b)
    {
        switch (returnType(a, b)) {
            case UNKNOWN:
            case DOUBLE: return new Double(a.doubleValue() + b.doubleValue());
            case FLOAT: return new Float(a.floatValue() + b.floatValue());
            case LONG: return new Long(a.longValue() + b.longValue());
            case INTEGER: return new Integer(a.intValue() + b.intValue());
            default: throw new Error("Unexpected number type.");
        }
    }

    public static boolean smaller(Number a, Number b)
    {
        switch (returnType(a, b)) {
            case UNKNOWN:
            case DOUBLE: return a.doubleValue() < b.doubleValue();
            case FLOAT: return a.floatValue() < b.floatValue();
            case LONG: return a.longValue() < b.longValue();
            case INTEGER: return a.intValue() < b.intValue();
            default: throw new Error("Unexpected number type.");
        }
    }

    public static Number times(Number a, Number b)
    {
        switch (returnType(a, b)) {
            case UNKNOWN:
            case DOUBLE: return new Double(a.doubleValue() * b.doubleValue());
            case FLOAT: return new Float(a.floatValue() * b.floatValue());
            case LONG: return new Long(a.longValue() * b.longValue());
            case INTEGER: return new Integer(a.intValue() * b.intValue());
            default: throw new Error("Unexpected number type.");
        }
    }

    public static Number trunc(Number a)
    {
        switch (getType(a)) {
            case UNKNOWN:
            case DOUBLE: return new Long(a.longValue());
            case FLOAT: return new Integer(a.intValue());
            case LONG:
            case INTEGER: return a;
            default: throw new Error("Unexpected number type.");
        }
    }

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final int UNKNOWN = 0;
    private static final int INTEGER = 1;
    private static final int LONG    = 2;
    private static final int FLOAT   = 3;
    private static final int DOUBLE  = 4;

    private static int getType(Number number)
    {
        if (number.getClass() == Double.class) return DOUBLE;
        else if (number.getClass() == Float.class) return FLOAT;
        else if (number.getClass() == Long.class) return LONG;
        else if (number.getClass() == Integer.class) return INTEGER;
        else if (number.getClass() == Short.class) return INTEGER;
        else if (number.getClass() == Byte.class) return INTEGER;
        else return UNKNOWN;
    }

    private static int returnType(Number a, Number b)
    {
        return Math.max(getType(a), getType(b));
    }

    private NumberOp()
    {
    }

}
