package ch.unibe.piccola.code;

import java.io.IOException;
import java.util.Arrays;
import ch.unibe.piccola.source.SourceRef;

/**
 * Represents a sequence of Piccola machine instructions.
 *
 * @author   Stefan Kneubuehl
 * @version  3.4c (Apr 02, 2003)
**/
public final class Code
{

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Compares the specified object with this code for equality. Returns
     * <code>true</code> if and only if the specified object is of the same
     * class and contains the same instructions in in the same order.
     *
     * @param   object the object to be compared for equality with this
     *          instruction list.
     * @return  <code>true</code> if the specified object is equal to this
     *          instruction list, <code>false</code> otherwise.
    **/
    public final boolean equals(Object object)
    {
        if (null == object) {
            return false;
        }
        if (!getClass().equals(object.getClass())) {
            return false;
        }
        return Arrays.equals(m_code, ((Code) object).m_code);
    }

    /**
     * Returns the instruction at the specified position in this sequence.
     *
     * @param   index index of element to return.
     * @return  the instruction at the specified position in this sequence.
    *  @throws  ArrayIndexOutOfBoundsException - if the index is out of range
    *           (<code>index < 0 || index >= size()</code>).
    **/
    public final OpCode get(int index)
    {
        return (OpCode) m_code[index];
    }

    /**
     * Returns the source code reference for this instruction sequence.
     *
     * @return  the source code reference for this instruction sequence.
    **/
    public final SourceRef getSourceRef()
    {
        return m_sourceRef;
    }

    /**
     * Returns the hash code for this instruction sequence.
     *
     * @return the hash code for this instruction sequence.
    **/
    public final int hashCode()
    {
        return m_code.hashCode();
    }

    /**
     * Returns a new code frame ready to execute this instruction sequence.
     *
     * @return a new code frame ready to execute this instruction sequence.
    **/
    public final CodeFrame newFrame()
    {
        return new CodeFrame(this);
    }

    /**
     * Returns the number of elements in this sequence. If this list contains
     * more than <code>Integer.MAX_VALUE</code> elements, returns
     * <code>Integer.MAX_VALUE</code>.
     *
     * @return  the number of elements in this sequence.
    **/
    public final int size()
    {
        return m_code.length;
    }

    /**
     * Returns a string describing this code. The format is <code>"Code"</code>
     * followed by a comma-separated list of the instructions in parantheses.
     *
     * @return a string describing this instruction sequence.
    **/
    public final String toString()
    {
        String result = "Code(";
        if (0 < m_code.length) {
            result += m_code[0];
        }
        for (int i = 1; i < m_code.length; ++i) {
            result += ", " + m_code[i];
        }
        return result + "]";
    }

    // ------------------------------------------------------------------------
    // default
    // ------------------------------------------------------------------------

    /**
     * Creates a new sequence of Piccola machine instructions.
     *
     * @param code the array of instructions to be represented by this
     *        sequence.
    **/
    Code(OpCode[] code)
    {
        m_code = code;
        if (size() == 0) {
            m_sourceRef = null;
        }
        else {
            SourceRef begin = get(0).getSourceRef();
            SourceRef end = get(size() - 1).getSourceRef();
            m_sourceRef = (null != begin && null != end) ?
                new SourceRef(begin, end) : null;
        }
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private final OpCode[] m_code;

    private final SourceRef m_sourceRef;

}