package ch.unibe.piccola.code;

import java.util.ArrayList;
import ch.unibe.piccola.source.SourceRef;

/**
 * <p>This class is used to build Piccola code. Use the <code>emit...</code>
 * methods to add opcodes to an internal list of opcodes. Then use
 * </code>code</code> to create the corresponding Piccola code. For example,
 * the following code builds Piccola code for a service that returns the form
 * <i>label = ()</i>.</p>
 * <p><code>CodeFactory cf = new CodeFactory();<br/>
 * cf.emitEmpty(null);<br/>
 * cf.emitBind(null, "label");<br/>
 * cf.emitExit();<br/>
 * Code code = cf.code();
 * </code></p>
 *
 * @see ch.unibe.piccola.code.IVM
 *
 * @author   Stefan Kneubuehl
 * @version  3.6c (May 11, 2003)
**/
public class CodeFactory
{

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Creates a new code factory.
    **/
    public CodeFactory()
    {
        m_commands = new ArrayList();
    }

    /**
     * Resets the CodeFactory, i.e. clears the list of opcodes.
    **/
    public final void clear()
    {
        m_commands.clear();
    }

    /**
     * Returns a new Code object representing the emitted opcodes, i.e. the
     * current list of opcodes.
     *
     * @return a new Code object representing the emitted opcodes.
    **/
    public final Code code()
    {
        return new Code(
            (OpCode[]) m_commands.toArray(new OpCode[m_commands.size()])
        );
    }

    /**
     * Appends an <b>apply</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitApply(SourceRef sourceRef)
    {
        addCommand(new OpApply(sourceRef));
    }

    /**
     * Appends a <b>bind</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
     * @param  label the label parameter for the <b>bind</b> opcode.
    **/
    public final void emitBind(SourceRef sourceRef, String label)
    {
        addCommand(new OpBind(sourceRef, label));
    }

    /**
     * Appends an <b>empty</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitEmpty(SourceRef sourceRef)
    {
        addCommand(new OpEmpty(sourceRef));
    }

    /**
     * Appends an <b>exit</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitExit(SourceRef sourceRef)
    {
        addCommand(new OpExit(sourceRef));
    }

    /**
     * Appends an <b>extend</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitExtend(SourceRef sourceRef)
    {
        addCommand(new OpExtend(sourceRef));
    }

    /**
     * Appends a <b>hide</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
     * @param  label the label parameter for the <b>hide</b> opcode.
    **/
    public final void emitHide(SourceRef sourceRef, String label)
    {
        addCommand(new OpHide(sourceRef, label));
    }

    /**
     * Appends a <b>javaCall</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
     * @param  service the service to be executed by this java call.
     * @param  target the target of this java call.
    **/
    public final void emitJavaCall(SourceRef sourceRef, JavaOp javaOp,
                                   Object target)
    {
        addCommand(new OpJavaCall(sourceRef, javaOp, target));
    }

    /**
     * Appends a <b>javaObject</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
     * @param  object the object parameter for the <b>javaObject</b> opcode.
    **/
    public final void emitJavaObject(SourceRef sourceRef, Object object)
    {
        addCommand(new OpJavaObject(sourceRef, object));
    }

    /**
     * Appends a <b>lookup</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
     * @param  label the label parameter for the <b>lookup</b> opcode.
    **/
    public final void emitLookup(SourceRef sourceRef, String label)
    {
        addCommand(new OpLookup(sourceRef, label));
    }

    /**
     * Appends an <b>meta</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitMeta(SourceRef sourceRef)
    {
        addCommand(new OpMeta(sourceRef));
    }

    /**
     * Appends a <b>popContext</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitPopContext(SourceRef sourceRef)
    {
        addCommand(new OpPopContext(sourceRef));
    }

    /**
     * Appends a <b>popValue</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitPopValue(SourceRef sourceRef)
    {
        addCommand(new OpPopValue(sourceRef));
    }

    /**
     * Appends a <b>project</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
     * @param  label the label parameter for the <b>project</b> opcode.
    **/
    public final void emitProject(SourceRef sourceRef, String label)
    {
        addCommand(new OpProject(sourceRef, label));
    }

    /**
     * Appends a <b>pushContext</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitPushContext(SourceRef sourceRef)
    {
        addCommand(new OpPushContext(sourceRef));
    }

    /**
     * Appends a <b>receive</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitReceive(SourceRef sourceRef, Object channel)
    {
        addCommand(new OpReceive(sourceRef, channel));
    }

    /**
     * Appends a <b>static</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitRoot(SourceRef sourceRef)
    {
        addCommand(new OpRoot(sourceRef));
    }

    /**
     * Appends a <b>send</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitSend(SourceRef sourceRef, Object channel)
    {
        addCommand(new OpSend(sourceRef, channel));
    }

    /**
     * Appends a <b>service</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
     * @param  code the code parameter for this <b>service</b> opcode.
    **/
    public final void emitService(SourceRef sourceRef, Code code)
    {
        addCommand(new OpService(sourceRef, code));
    }

    /**
     * Appends a <b>variable</b> opcode to the current list of opcodes.
     *
     * @param  sourceRef the reference to the source code which caused this
     *         opcode to be emitted.
    **/
    public final void emitVariable(SourceRef sourceRef)
    {
        addCommand(new OpVariable(sourceRef));
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private ArrayList m_commands;

    private final void addCommand(OpCode opCode)
    {
        m_commands.add(opCode);
    }

}

