package ch.unibe.piccola.compiler;

import ch.unibe.piccola.syntax.*;
import ch.unibe.piccola.source.SourceRef;

/**
 * This class implements a syntax tree visitor that processes a Piccola syntax
 * tree. The Simplifier takes a syntax tree as input, and replaces nodes of the
 * following kinds with other nodes:
 * <ul>
 *   <li>Collection</li>
 *   <li>CompoundBinding</li>
 *   <li>CurriedAbstraction</li>
 *   <li>InfixApplication</li>
 *   <li>NamedAbstraction</li>
 *   <li>PrefixApplication</li>
 *   <li>Quote</li>
 * </ul>
 * The simplification is an idempotent operation, i.e.
 * <code>
 *   Simplifier.apply(Simplifier.apply(node)) == Simplifier.apply(node)
 * </code>
 *
 * @author   Stefan Kneubuehl
 * @version  3.4c (Sep 30, 2001)
**/
public final class Simplifier extends NodeVisitor
{

    // ------------------------------------------------------------------------
    // public static
    // ------------------------------------------------------------------------

    public static final Node apply(Node node)
    {
        return INSTANCE.visit(node);
    }

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String LABEL_DEFAULTOP = Syntax.labelDefaultOp();

    private static final String LABEL_ADD = Syntax.labelAdd();

    private static final String LABEL_L = Syntax.labelL();

    private static final String LABEL_R = Syntax.labelR();

    private static final Simplifier INSTANCE = new Simplifier();

    /**
     * Input : x, F1, F2
     * Output: root = (root, F1), (x = x), F2
    **/
    private static final Node makeBang(SourceRef sr, String label, Node F1, Node F2)
    {
        // x = x
        Node F = sf.newBinding(sr, sf.newLabel(sr, label), label);
        // (x = x), F2
        F = sf.newExtension(sr, F, F2);
        // root, F1
        Node G = sf.newExtension(sr, sf.newRoot(sr), F1);
        // root = (root, F1), (x = x), F2
        return sf.newSandbox(sr, G, F);
    }

    /**
     * Input: labels, F1, F2
     * Output: simplification of l1.l2.l3 = F1, F2
    **/
    private static final Node makeNestedBinding(SourceRef sr, boolean fp,
                                                String[] labels, Node form, Node F2)
    {
        int first = (labels[0].equals("root")) ? 1 : 0;
        int i = labels.length - 1;
        while (i > first) {
            if (i == labels.length - 1 && fp) {
                form = sf.newFixedPoint(sr, form, labels[i]);
            }
            else {
                form = sf.newBinding(sr, form, labels[i]);
            }
            Node proj = sf.newLabel(sr, labels[first]);
            for (int j = first+1; j < i; ++j) {
                proj = sf.newProjection(sr, proj, labels[j]);
            }
            form = sf.newExtension(sr, proj, form);
            --i;
        }
        if (labels[0].equals("root")) {
            return sf.newSandbox(sr,
                sf.newExtension(sr, sf.newRoot(sr), sf.newBinding(sr, form, labels[1])), F2
            );
        }
        else {
            if (labels.length == 1 && fp) {
                form = sf.newFixedPoint(sr, form, labels[0]);
            }
            else {
                form = sf.newBinding(sr, form, labels[0]);
            }
            return (F2 instanceof Empty) ? form : makeBang(sr, labels[0], form, F2);
        }
    }

    /**
     * (abb-nest)
     *
     * L.I = F      ==>  L = (L, I = F)
     * def L.I = F  ==>  L = (L, def I = mu[F])
    **/
    private static final Node makeBinding(SourceRef sr, boolean fp,
                                          String[] labels, Node form)
    {
        int i = labels.length - 1;
        while (i > 0) {
            if (i == labels.length - 1 && fp) {
                form = sf.newFixedPoint(sr, form, labels[i]);
            }
            else {
                form = sf.newBinding(sr, form, labels[i]);
            }
            Node proj = sf.newLabel(sr, labels[0]);
            for (int j = 1; j < i; ++j) {
                proj = sf.newProjection(sr, proj, labels[j]);
            }
            form = sf.newExtension(sr, proj, form);
            --i;
        }
        if (labels.length == 1 && fp) {
            return sf.newFixedPoint(sr, form, labels[0]);
        }
        else {
            return sf.newBinding(sr, form, labels[0]);
        }
    }

    /**
     * Input : x1, x2, ... , xn, F
     * Output: \x1: \x2: ... \xn: F
    **/
    private static final Node makeService(SourceRef sr, String[] params, Node form)
    {
        for (int i = params.length - 1; i >= 0; --i) {
            form = sf.newAbstraction(sr, form, params[i]);
        }
        return form;
    }

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Applies the following transformations:
     *     simplify.{}       = DefaultOp.{_} ()
     *     simplify.{ L, F } = (simplify.{L}).add(simplify(F))
    **/
    public final Object visitCollection(Collection node)
    {
        SourceRef sr = node.getSourceRef();
        Node result = sf.newLabel(sr, LABEL_DEFAULTOP);
        result = sf.newProjection(sr, result, node.getLabel());
        result = sf.newApplication(sr, result, sf.newEmpty(sr));
        Node[] children = node.getChildren();
        for (int i = 0; i < children.length; ++i) {
            result = sf.newProjection(sr, result, LABEL_ADD);
            result = sf.newApplication(sr, result, visit(children[i]));
        }
        return result;
    }

    /**
     * 1. nested labels
     * 2. continuation
     * 3. fixed point
    **/
    public final Object visitCompoundBinding(CompoundBinding node)
    {
        SourceRef sr = node.getSourceRef();
        Node F1 = visit(node.getLeftChild());
        Node F2 = visit(node.getRightChild());
        String[] labels = node.getLabels();

        return makeNestedBinding(sr, node.isFixedPoint(), labels, F1, F2);
    }

    /**
     * si[\x y z: F]  ==>  \x: \y: \z: si[F]
    **/
    public final Object visitCurriedAbstraction(CurriedAbstraction node)
    {
        return makeService(node.getSourceRef(), node.getParameters(),
                           visit(node.getChild()));
    }

    /**
     * si[F1 + F2]
     *   ==>
     * ( 'l = si[F1], ( _+_ = \r: DefaultOp._+_default l r, l ) )._+_ si[F2]
    **/
    public final Object visitInfixApplication(InfixApplication node)
    {
        SourceRef sr = node.getSourceRef();
        String opLabel = Syntax.labelInfixOp(node.getLabel());
        Node lLabel = sf.newLabel(sr, LABEL_L);
        Node rLabel = sf.newLabel(sr, LABEL_R);

        // L = F1
        Node form = sf.newBinding(sr, visit(node.getLeftChild()), LABEL_L);
        // DefaultOp
        Node result = sf.newLabel(sr, LABEL_DEFAULTOP);
        // DefaultOp._+_default
        result = sf.newProjection(sr, result, Syntax.labelInfixOpDefault(node.getLabel()));
        // DefaultOp._+_default L
        result = sf.newApplication(sr, result, lLabel);
        // DefaultOp._+_default L R
        result = sf.newApplication(sr, result, rLabel);
        // \R: DefaultOp._+_default L R
            result = sf.newAbstraction(sr, result, LABEL_R);
        // _+_ = \R: DefaultOp._+_default L R
        result = sf.newBinding(sr, result, opLabel);
        // ( _+_ = \R: DefaultOp._+_default L R, L )
        result = sf.newExtension(sr, result, lLabel);
        // 'L = F, ( _+_ = \R: DefaultOp._+_default L R, L )
        form = sf.newExtension(sr, sf.newRoot(sr), form);
        result = sf.newSandbox(sr, form, result);
        // ( 'L = F, ( _+_ = \R: DefaultOp._+_default L R, L ) )._+_
        result = sf.newProjection(sr, result, opLabel);
        // ( 'L = F, ( _+_ = \R: DefaultOp._+_default L R, L ) )._+_ F2
        return sf.newApplication(sr, result, visit(node.getRightChild()));
    }

    /**
     * 1. nested params (see curried abstraction)
     * 2. nested labels (see compound binding)
     * 3. continuation  (see compound binding)
     * 4. fixed point (see compound binding)
    **/
    public final Object visitNamedAbstraction(NamedAbstraction node)
    {
        SourceRef sr = node.getSourceRef();
        Node F1 = visit(node.getLeftChild());
        Node F2 = visit(node.getRightChild());
        String[] labels = node.getLabels();

        F1 = makeService(sr, node.getParameters(), F1);
        return makeNestedBinding(sr, node.isFixedPoint(), labels, F1, F2);
    }

    /**
     * si[- F]
     *    ==>
     * ( 'R = si[F], ( -_ = \: DefaultOp.-_default R, R ) ).-_ ()
    **/
    public final Object visitPrefixApplication(PrefixApplication node)
    {
        SourceRef sr = node.getSourceRef();
        String opLabel = Syntax.labelPrefixOp(node.getLabel());
        Node rLabel = sf.newLabel(sr, LABEL_R);

        // R = F
        Node form = sf.newBinding(sr, visit(node.getChild()), LABEL_R);
        // DefaultOp
        Node result = sf.newLabel(sr, LABEL_DEFAULTOP);
        // DefaultOp.-_default
        result = sf.newProjection(sr, result, Syntax.labelPrefixOpDefault(node.getLabel()));
        // DefaultOp.-_default R
        result = sf.newApplication(sr, result, rLabel);
        // {op} = \: DefaultOp.-_default R
        result = sf.newBinding(sr, sf.newAbstraction(sr, result, ""), opLabel);
        // ( {op} = \: DefaultOp.-_default R, R )
        result = sf.newExtension(sr, result, rLabel);
        // 'R = F, ( -_ = \: DefaultOp.-_default R, R )
        form = sf.newExtension(sr, sf.newRoot(sr), form);
        result = sf.newSandbox(sr, form, result);
        // ( 'R = F, ( -_ = \: DefaultOp.-_default R, R ) ).-_
        result = sf.newProjection(sr, result, opLabel);
        // ( 'R = F, ( -_ = \: DefaultOp.-_default R, R ) ).-_ ()
        return sf.newApplication(sr, result, sf.newEmpty(sr));
    }

    /**
     * Applies the following transformation:
     *     simplify.('E, F) = (root = (root, simplify.E), simplify.F)
    **/
    public final Object visitQuote(Quote node)
    {
        SourceRef sr = node.getSourceRef();
        Node F1 = visit(node.getLeftChild());
        Node F2 = visit(node.getRightChild());

        F1 = sf.newExtension(sr, sf.newRoot(sr), F1);
        return sf.newSandbox(sr, F1, F2);
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private Simplifier()
    {
    }

}

