package ch.unibe.piccola.compiler;

import junit.framework.TestCase;
import junit.framework.TestSuite;
import ch.unibe.piccola.parser.Parser;
import ch.unibe.piccola.parser.ParserException;
import ch.unibe.piccola.parser.ParserFactory;
import ch.unibe.piccola.source.StringSourceCode;
import ch.unibe.piccola.syntax.Node;

/**
 * @author   Stefan Kneubuehl
 * @version  3.4c (Oct 02, 2002)
**/
public class SimplifierTest extends TestCase {

    // ------------------------------------------------------------------------
    // test suite
    // ------------------------------------------------------------------------

    public SimplifierTest(String name)
    {
        super(name);
    }

    public static TestSuite suite()
    {
        return new TestSuite(SimplifierTest.class);
    }

    public static void main(String[] args)
    {
        junit.textui.TestRunner.run(suite());
    }

    // ------------------------------------------------------------------------
    // test environment
    // ------------------------------------------------------------------------

    public void setUp()
    {
        parser = ParserFactory.INSTANCE.newJavaCCParser();
    }

    private Parser parser;

    private Node parse(String code)
    {
        try {
            return parser.parse(new StringSourceCode(code));
        }
        catch (ParserException ex) {
            fail("parsing failed: while parsing \"" + code + "\" " + ex);
        }
        catch (Throwable ex) {
            fail("Unexpected exception: " + ex);
        }
        return null;
    }

    private Node siParse(String source)
    {
        return Simplifier.apply(parse(source));
    }

    private void assertSimplify(String expect, String get)
    {
        assertEquals(parse(expect), siParse(get));
    }

    // ------------------------------------------------------------------------
    // tests
    // ------------------------------------------------------------------------

    /**
     * Tests for collections
    **/
    public void testCollection()
    {
        assertSimplify("DefaultOp.{_}()", "{ }");
        assertSimplify("DefaultOp.{_}().add(x)", "{ x }");
        assertSimplify(
            "DefaultOp.{_}().add(x).add(a, b).add(y)",
            "{ x, (a, b), y }"
        );
    }

    /**
     * Tests for compound bindings
    **/
    public void testCompoundBinding()
    {
        assertSimplify("root = (root, a = ()), (a = a), b", "a = (), b");
        assertSimplify("a = (a, b = c)", "a.b = c");
        assertSimplify(
            "root = (root, a = (a, b = c)), ((a = a), form)",
            "a.b = c, form"
        );
        assertSimplify("a, f = (f, g = (f.g, y = x))", "a, f.g.y = x");
        assertSimplify("a = (a, def b = ())", "def a.b = ()");
        assertSimplify("root = (root, def a = ()), (a = a), f", "def a = (), f");
        assertSimplify("root = (root, label = ()), x", "root.label = (), x");
        assertSimplify("root = (root, a = (a, b = ())), x", "root.a.b = (), x");
        assertSimplify("root = (root, d = (d, e = z)), (d = d), x", "d.e = z, x");
    }

    /**
     * Tests for curried abstraction
    **/
    public void testCurriedAbstraction()
    {
        assertSimplify("\\x: \\y: \\: test", "\\x y (): test");
    }

    /**
     * Tests for infix application
    **/
    public void testInfix()
    {
        assertSimplify("((root = (root, L = a), " +
                       "((_+_ = \\R: DefaultOp._+_default L R), L))._+_) b",
            "a + b");
    }

    /**
     * Tests for label
    **/
    public void testLabel()
    {
        assertSimplify("label", "label");
    }

    /**
     * Tests for named abstraction
    **/
    public void testNamedAbstraction()
    {
        assertSimplify("a = (a, b = (\\: ()))", "a.b: ()");
        assertSimplify("root = (root, b = (\\: ()))", "root.b: ()");
    }

    /**
     * Tests for prefix application
    **/
    public void testPrefix()
    {
        assertSimplify("((root = (root, (R = form)), " +
                       "(-_ = (\\: (DefaultOp.-_default R))), R).-_) ()",
            "-form");
    }

    /**
     * Tests for quote
    **/
    public void testQuote()
    {
        assertSimplify("root = (root, x), ()", "'x");
        assertSimplify("root = (root, x), x", "'x, x");
        assertSimplify("root = (root, f = \\:()), f", "'f: (), f");
        assertSimplify("root = (root, (root = (root, f()), ())), x",
            "''f(), x");
    }

}