package ch.unibe.piccola.parser.javacc;

/**
 * Parses a literal.
 *
 * @author   Franz Achermann, Stefan Kneubuehl
 * @version  3.4c (Jul 19, 2002)
**/
public final class LiteralParser
{

    // ------------------------------------------------------------------------
    // public static
    // ------------------------------------------------------------------------

    /**
     * Returns an object representing the specified literal. The returned
     * object is an instance of one of the following classes:
     * <ul>
     *   <li>java.lang.Integer</li>
     *   <li>java.lang.Float</li>
     *   <li>java.lang.Double</li>
     *   <li>java.lang.Long</li>
     *   <li>java.lang.String</li>
     * </ul>
     * Returns null, if the literal has been initialized with an invalid
     * literal string.
     *
     * @return   an object representing the specified literal.
    **/
    public static final Object parse(String literal)
    {
        Object result = null;
        if (literal.startsWith(SINGLE)) {
            return processString(literal);
        }
        try {
            return new Integer(Integer.parseInt(literal));
        }
        catch (NumberFormatException ex) { }
        try {
            return new Long(Long.parseLong(literal));
        }
        catch (NumberFormatException ex) { }
        try {
            return new Float(Float.parseFloat(literal));
        }
        catch (NumberFormatException ex) { }
        try {
            return new Double(Double.parseDouble(literal));
        }
        catch (NumberFormatException ex) { }
        return null;
    }

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final char ESC = '\\';

    private static final char ESC_TAB = 't';

    private static final char ESC_NL = 'n';

    private static final char TAB = '\t';

    private static final char NL = '\n';

    private static final char CR = '\r';

    private static final char QUOTE = '\"';

    private static final char ZERO = '0';

    private static final char NINE = '9';

    private static final String SINGLE = "\"";

    private static final String TRIPLE = "\"\"\"";

    /**
     * Processes a string literal by replacing escape sequences with the
     * corresponding character and by trimming leading and trailing double
     * quotes ('\"'). The following escape sequences are replaced:
     * '\n', '\t', '\"', '\\', '\ooo' where ooo is an octal number.
     * Leaves triple quoted string unchanged.
    **/
    private static final String processString(String input)
    {
        if (input.startsWith(TRIPLE)) {
            return input.substring(3, input.length() - 3);
        }
        int length = input.length() - 1; // omit trailing "
        int onull = (int) ZERO;
        StringBuffer result = new StringBuffer(length);
        int pos = 1; // omit leading "
        char curr;
        char next;
        while (pos < length) {
            curr = input.charAt(pos++);
            if (curr == ESC) {
                next = input.charAt(pos++);
                switch (next) {
                case ESC_NL:
                    curr = NL;
                    break;
                case ESC_TAB:
                    curr = TAB;
                    break;
                case ESC:
                    curr = ESC;
                    break;
                case QUOTE:
                    curr = QUOTE;
                    break;
                case NL:
                case CR:
                    if ((input.charAt(pos) == NL) ||
                        (input.charAt(pos) == CR)) {
                        ++pos;
                    }
                    curr = input.charAt(pos++);
                    break;
                default:
                    if (next >= ZERO && next <= NINE &&
                        pos+1 < length &&
                        input.charAt(pos) >= ZERO &&
                        input.charAt(pos) <= NINE &&
                        input.charAt(pos+1) >= ZERO &&
                        input.charAt(pos+1) <= NINE) {

                        int octal = ((int) next - onull) * 80;
                        octal += ((int) input.charAt(pos++) - onull) * 8;
                        octal += (int) input.charAt(pos++) - onull ;
                        curr = (char) octal;
                    } else {
                        curr = next;
                    }
                    break;
                } // select
            } // if
            result.append(curr);
        }
        return result.toString();
    }

}