package ch.unibe.piccola.parser.javacc;
/* File: Scanner.java
 * Author: F. Achermann acherman@iam.unibe.ch
 * Intend: Scanner for JavaCC Parser
 * History: 22.9.98 started
 *          2.2001  code duplication to treat LPAREN and RPAREN as left and right
 *
 * delegates all calls to ParserTokenManager, but from time to time, inserts
 * and or moves Tokens
 */

import java.util.*;


/** a Wrapper around the generated Tokenmanager to insert dedent and indent according
 *  to newline changes.
 */
public class Scanner implements ParserImplConstants {
	private final ParserImplTokenManager m;
	private Stack next = null;
	private ASCII_CharStream is;

	protected Token lastToken;
	protected Token nextToken;

	private Stack indentation;
	private int indentationLevel;
	private int nextEnforcedIndentationLevel;

	private Scanner() { m = null;  }

	public Scanner(ASCII_CharStream stream) {
		m = new ParserImplTokenManager(stream);
		is = stream;
		indentationLevel = 0;
		nextEnforcedIndentationLevel = 0;
		next = new Stack();
		lastToken = null;
		indentation = new Stack();
		Token t = newToken(new Token(), LEFT, "(");
		t.indentationLevel = 1;
		indentation.push(t);
	}

	public Scanner(ASCII_CharStream stream, int lexState) {
		m = new ParserImplTokenManager(stream, lexState);
		is = stream;
		indentationLevel = 0;
		nextEnforcedIndentationLevel = 0;
		next = new Stack();
		lastToken = null;
		indentation = new Stack();
		Token t = newToken(new Token(), LEFT, "(");
		t.indentationLevel = 1;
		indentation.push(t);
	}

	public void ReInit(ASCII_CharStream stream) {
		m.ReInit(stream);
		is = stream;
		indentationLevel = 0;
		nextEnforcedIndentationLevel = 0;
		next = new Stack();
		lastToken = null;
		indentation = new Stack();
		Token t = newToken(new Token(), LEFT, "(");
		t.indentationLevel = 1;
		indentation.push(t);
	}



 /** return a comma token
   */
  protected Token newToken(Token current, int kind, String image) {
      Token t = new Token();
      t.kind = kind;
      t.m_beginPos = current.m_beginPos;
      t.beginLine = current.beginLine;
      t.beginColumn = current.beginColumn;
      t.endLine = current.beginLine;
      t.endColumn = current.beginColumn;
      t.image = image;
      return t;
  }

 /** return a comma token
   */
  protected Token newToken(Token current) {
      Token t = new Token();
      t.kind = current.kind;
      t.m_beginPos = current.m_beginPos;
      t.beginLine = current.beginLine;
      t.beginColumn = current.beginColumn;
      t.endLine = current.beginLine;
      t.endColumn = current.beginColumn;
      t.image = current.image;
      return t;
  }




  protected void indent(Token current, Token la) {
      if (!(
          (current.kind == LEFT) ||
          (current.kind == LPAREN) ||
          (la.kind == DOT) ||
          false)) {
              Token t = newToken(current, LEFT, "(");
              t.indentationLevel = indentationLevel;
              next.push(t);
              pushOpening(t);
          }
  }

  protected void newLine(Token current, Token la) {
      if (!(
          (current.kind == LEFT) ||
          (current.kind == LPAREN) ||
          (current.kind == COMMA) ||
          (current.kind == OP1) ||
          (current.kind == OP2) ||
          (current.kind == OP3) ||
          (current.kind == OP4) ||
          (la.kind == DOT) ||
          false)) {
              Token t = newToken(current, COMMA, ",");
              next.push(t);
          }
  }

  protected void dedent(Token current, Token la) {
      if (!(
          (la.kind == RIGHT) ||
          (la.kind == RPAREN) ||
          false)) {
              dedentPossible(la);
              if (la.kind != DOT) {
                  Token t = newToken(current, COMMA, ",");
                  next.insertElementAt(t, 0);
              }
          }
  }


    protected void enforceNextIndentation() {
        int s = indentation.size();
        // System.out.println("enforce: " + indentation);
        while (s > 0) {
            Token t = (Token)indentation.elementAt(--s);
            if (t.kind == LEFT && t.indentationLevel > 0) {
                nextEnforcedIndentationLevel = t.indentationLevel;
                // System.out.println("next line must start on: " + nextEnforcedIndentationLevel);
                return;
            }
        }
        nextEnforcedIndentationLevel = 1;
        // System.out.println("next line must start on (stack empty): " + nextEnforcedIndentationLevel);
    }


  protected void dedentPossible(Token nextToken) {
        // search for topmost token in stack with indentationLevel.
        // all above must be indent.
        // System.out.println("dedent Possible: " + indentation);
        int s = indentation.size();
        while (s > 0) {
            Token t = (Token)indentation.elementAt(--s);
            // System.out.println("considering token: " + t + " next token " + nextToken);
            if ((t.kind == LEFT && t.indentationLevel == 0) ||
                 t.kind == LPAREN ||
                 t.kind == TIGHTLPAREN ||
                 t.kind == TIGHTLEFT) {
                // System.out.println("No dedent possible: ");
                break;
            }
            if (t.kind != LEFT || t.indentationLevel < nextToken.beginColumn) {
                throw badIndentation(nextToken);
            } else {
                if (t.indentationLevel == nextToken.beginColumn) {
                    // System.out.println("dedent cleaned up (next token): " + nextToken.beginColumn);
                    // System.out.println("dedent cleaned up: " + indentation);
                    // System.out.println("Next tokens: " + next);
                    break;
                }
                indentation.pop();
                next.insertElementAt(newToken(nextToken, RIGHT, ")"), 0);
            }
        }
        // System.out.println("Stack now is: " + indentation);
  }




  protected void pushOpening(Token t) {
      switch (t.kind) {
      case LEFT:
      case LPAREN: {
          indentation.push(t);
          break;
      }
      default: {}
      }
  }

  /** [ matches with ]
   *  (              )
   *  (n             )n where n is in the image as well as the endColumn
   *  of the tokens
   */
  protected static boolean matches(Token left, Token right) {
      if ((left.kind == LPAREN || left.kind == TIGHTLPAREN) &&
           right.kind == RPAREN) return true;
      if ((left.kind == LEFT || left.kind == TIGHTLEFT) &&
          right.kind == RIGHT && left.indentationLevel == right.indentationLevel)
            return true;
      return false;

  }

  protected TokenMgrError badIndentation(Token t) {
      return new TokenMgrError("Token with bad indentation: " + t + " line: " + t.beginLine + " column: " + t.beginColumn, 0);

  }


  /** returns null when there was an error
   *  client should raise exception
   */
  protected Token insertClosing(Token t) {
      switch (t.kind) {
      case RIGHT:
      case RPAREN: {
          // System.out.println("Closing with : " + t);
          // System.out.println("next stack: " + next);
          // System.out.println("Indentation stack: " + indentation);
          while (indentation.size()>1) {
              Token top = (Token)indentation.pop();
              // System.out.println("Top token: " + top);
              if (matches(top, t)) {
                  // System.out.println("It matches: " + top + " " + t);
                  next.insertElementAt(t, 0);
                  // System.out.println("The stack of next tokens: " + next);
                  // enforceNextIndentation();;
                  return (Token)next.pop();
              } else {
                  if (top.kind == LEFT && top.indentationLevel > 0) {
                      // it was an indent
                      enforceNextIndentation();
                      if (nextEnforcedIndentationLevel > 0 && nextToken.beginLine != t.beginLine && nextToken.beginColumn > nextEnforcedIndentationLevel) {
                          throw badIndentation(nextToken);
                      }
                      next.push(newToken(t, RIGHT, ")"));
                  } else {
                      // System.out.println("Wrong closing bracket: " + t);
                      // cannot close a bracket or user parentheses.
                      // this will cause a parse exception;
                      return null;
                      // return t;
                  }
              }
          }
          // this will cause a parse exception.  Token t is on an ivalid indentation
          // System.out.println("Stack is empty:");
          return null;
          // throw badIndentation(t);
          // return t;
      }
      default: { return t; }
      }
  }



	/** the public getnext mehtod
   *  if the emited token is LEFT we return a TIGHTLEFT
   *  but keep LEFT as lastEmitedToken
   */
	public final Token getNextToken() {

      // System.out.println("Start: " + nextToken + " next: " + next);

      // tokens in the next stack were inserted by a previous getNextToken
      // call. they are returned first.
      if (!next.isEmpty()) {
          return (Token)next.pop();
      }

      // nextToken is the lookahead token,
      if (nextToken == null) {
          nextToken = m.getNextToken();
          indentationLevel = nextToken.beginColumn;
          // System.out.println("lookahead received token: " + nextToken);
      }


      Token currentToken = nextToken;
      nextToken = m.getNextToken();
      // System.out.println("read token from stream: " + nextToken);

      pushOpening(currentToken);
      Token n = insertClosing(currentToken);
      if (n == null) {
          System.out.println("Error: token: " + currentToken);
          System.out.println("Error: token: " + nextToken);
          // the user typed a closing bracket, parentheses: the parser will complain
          n = currentToken;
      }
      currentToken = n;


      if (nextToken.kind != EOF) {
           if (nextToken.beginLine != currentToken.beginLine) {
              if (nextEnforcedIndentationLevel > 0 &&
                  nextToken.beginColumn > nextEnforcedIndentationLevel) {
                      throw badIndentation(nextToken);
              } else {
                  nextEnforcedIndentationLevel = 0;
              }

              if (indentationLevel < nextToken.beginColumn) {
                  indentationLevel = nextToken.beginColumn;
                  indent(currentToken, nextToken);
              } else if (indentationLevel == nextToken.beginColumn) {
                  newLine(currentToken, nextToken);
              } else {
                  dedent(currentToken, nextToken);
                  indentationLevel = nextToken.beginColumn;
              }
          }
      }


      // change left, lparen to its tight versions.
      if (lastToken != null &&
          lastToken.endLine == currentToken.beginLine &&
          lastToken.endColumn+1 == currentToken.beginColumn) {
          if (currentToken.kind == LEFT) {
              currentToken.kind = TIGHTLEFT;
          } else if (currentToken.kind == LPAREN) {
              currentToken.kind = TIGHTLPAREN;
          }
      }


      if (currentToken.kind == EOF && indentation.size()>1) {
            Token t = (Token)indentation.pop();
            if (t.indentationLevel > 0) {
	            currentToken = newToken(currentToken, RIGHT, ")");
	        }
      }

      lastToken = currentToken;
      return currentToken;
	}


}
