package ch.unibe.piccola.vm;

import ch.unibe.piccola.code.Code;
import ch.unibe.piccola.code.CodeFrame;
import ch.unibe.piccola.code.OpCode;
import ch.unibe.piccola.code.IVM;
import ch.unibe.piccola.source.SourceRef;
import ch.unibe.piccola.syntax.Syntax;

/**
 * This class implements an agent which can communicate with other agents over
 * channels.
 *
 * @author   Stefan Kneubuehl
 * @version  3.4c (Dec 02, 2002)
**/
public abstract class Agent implements IVM
{

    // ------------------------------------------------------------------------
    // public static
    // ------------------------------------------------------------------------

    public static final int INITIALIZED = 0;

    public static final int RUNNING = 1;

    public static final int BLOCKED = 2;

    public static final int STOPPED = 3;

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String LABEL_DYNAMIC = Syntax.labelDynamic();

    private static int NEXT_ID = 0;

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    public final Form evaluate()
    {
        setState(RUNNING);
        while (RUNNING == getState()) {
            mainLoop();
        }
        return getValue();
    }

    public final Form getContext()
    {
        return m_contextStack.peek();
    }

    public final Form getDynamic()
    {
        return getContext().project(LABEL_DYNAMIC);
    }

    public final int getId()
    {
        return m_id;
    }

    public SourceRef getSourceRef()
    {
        return (null == m_opCode) ? null : m_opCode.getSourceRef();
    }

    public final synchronized int getState()
    {
        return m_state;
    }

    public final Form getValue()
    {
        return m_valueStack.peek();
    }

    public final void start()
    {
        if (INITIALIZED != getState()) {
            throw new Error("Invalid agent state in start.");
        }
        if (null != m_observer) {
            m_observer.onStart(this);
        }
        setState(RUNNING);
        m_pool.requestThread(this);
    }

    public final void stop()
    {
        if (RUNNING != getState()) {
            throw new Error("Invalid agent state in stop.");
        }
        if (null != m_observer) {
            m_observer.onStop(this);
        }
        setState(STOPPED);
    }

    public final void waitUntilStopped()
    {
        while (STOPPED != getState()) {
            try {
                Thread.sleep(200);
            }
            catch (InterruptedException ex) { }
        }
    }

    // ------------------------------------------------------------------------
    // default
    // ------------------------------------------------------------------------

    final void block()
    {
        if (RUNNING != getState()) {
            throw new Error("Invalid agent state in block.");
        }
        if (null != m_observer) {
            m_observer.onBlock(this);
        }
        setState(BLOCKED);
    }

    public void mainLoop()
    {
        try {
            if (null != m_currentFrame) {
                m_opCode = m_currentFrame.next();
                if (null != m_observer) {
                    m_observer.beforeExecute(m_opCode);
                }
                m_opCode.execute(this);
                if (null != m_observer) {
                    m_observer.afterExecute(m_opCode);
                }
            }
            else {
                stop();
            }
        }
        catch (AgentException ex) {
            handleException(ex);
        }
    }

    final int sizeCode()
    {
        return m_codeStack.size();
    }

    final int sizeContext()
    {
        return m_contextStack.size();
    }

    final int sizeValue()
    {
        return m_valueStack.size();
    }

    final void popCode()
    {
        m_currentFrame = m_codeStack.pop();
    }

    final Form popContext()
    {
        return m_contextStack.pop();
    }

    final Form popValue()
    {
        return m_valueStack.pop();
    }

    final void pushCode(Code value)
    {
        m_codeStack.push(m_currentFrame);
        m_currentFrame = value.newFrame();
    }

    final void pushContext(Form value)
    {
        m_contextStack.push(value);
    }

    void pushValue(Form value)
    {
        m_valueStack.push(value);
    }

    final void unblock()
    {
        if (BLOCKED != getState()) {
            throw new Error("Invalid agent state in unblock.");
        }
        if (null != m_observer) {
            m_observer.onUnblock(this);
        }
        setState(RUNNING);
        m_pool.requestThread(this);
    }

    // ------------------------------------------------------------------------
    // protected
    // ------------------------------------------------------------------------

    protected Agent(AgentPool pool)
    {
        m_codeStack = new CodeFrameStack();
        m_contextStack = new FormStack();
        m_id = NEXT_ID++;
        m_observer = pool.newAgentObserver();
        m_pool = pool;
        m_state = INITIALIZED;
        m_valueStack = new FormStack();

    }

    protected SourceRef[] stackTrace()
    {
        return m_codeStack.stackTrace();
    }

    protected void handleException(AgentException ex)
    {
        System.out.println(ex);
        m_pool.abort();
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private final CodeFrameStack m_codeStack;

    private final FormStack m_contextStack;

    private CodeFrame m_currentFrame;

    private final int m_id;

    private IAgentObserver m_observer;

    private OpCode m_opCode;

    private AgentPool m_pool;

    private int m_state;

    protected final FormStack m_valueStack;

    private synchronized void setState(int state)
    {
        m_state = state;
    }

}

