package ch.unibe.piccola.vm;

import java.util.Iterator;
import java.util.SortedMap;
import java.util.TreeMap;
import ch.unibe.piccola.code.Code;
import ch.unibe.piccola.syntax.Syntax;
import ch.unibe.piccola.type.ContractualType;

/**
 * @author   Stefan Kneubuehl
 * @version  3.6c (Apr 23, 2003)
**/
public abstract class DefaultForm extends Form {

    // ------------------------------------------------------------------------
    // private static
    // ------------------------------------------------------------------------

    private static final String LABEL_ARG = Syntax.labelArg();

    private static final String LABEL_EQUALS = Syntax.labelEquals();

    private static final String LABEL_PEER = Syntax.labelPeer();

    private static final String LABEL_HASH = Syntax.labelHash();

    private static final String LABEL_TOSTRING = Syntax.labelToString();

    // ------------------------------------------------------------------------
    // public
    // ------------------------------------------------------------------------

    /**
     * Binds this form to the specified label.
     *
     * @param  label the label this form will be bound to.
     *
     * @return this form bound to the specified label, i.e.
     *         <i>(label = this)</i>.
    **/
    public final Form bind(String label)
    {
	    return m_factory.bindingForm(label, this);
    }

    /**
     * Downs this form.
     *
     * @return the object representing the downed form.
    **/
    public Object down()
    {
        return projectDef(LABEL_PEER, this).unwrap();
    }

    /**
     * Compares this form to the specified object.
     *
     * @param  object the object to be compared with this form.
     *
     * @return <code>true</code> if this form is equal to object,
     *         <code>false</code> otherwise.
    **/
    public boolean equals(Object object)
    {
        if (object instanceof Form) {
            Form other = (Form) object;
            Form op = projectDef(LABEL_EQUALS, null);
            if (null != op) {
                Object result = op.service().evaluate(other.bind(LABEL_ARG)).down();
                if (result instanceof Boolean) {
                    return ((Boolean) result).booleanValue();
                }
                throw new AgentException(
                    "Equality operator does not return boolean value."
                );
            }
            return runtimeEquals(other);
        }
        return false;
    }

    /**
     * Checks whether the specified label is bound by this form.
     *
     * @param  label the label for which the existence will be checked.
     *
     * @return <code>true</code> if this form binds the label,
     *         <code>false</code> otherwise.
    **/
    public abstract boolean exists(String label);

    /**
     * Extends this form with the specified form.
    **/
    public Form extend(Form form)
    {
        return form.isEmpty() ? this : m_factory.extensionForm(this, form);
    }

    /**
     * Returns a sorted map containing all bindings of this form. The keys of
     * the map are the labels represented as String object, the values are the
     * corresponding Forms.
     *
     * @return a SortedMap mapping the labels of this form to their values.
    **/
    public final SortedMap getBindings()
    {
        if (null == m_bindings) {
            m_bindings = buildBindings();
        }
        return m_bindings;
    }

    public Service service()
    {
        throw new AgentException("Apply failed. Form is not a service.");
    }

    /**
     * Returns the hash code of this form.
     *
     * @return the hash code of this form.
    **/
    public final int hashCode()
    {
        Form op = projectDef(LABEL_HASH, null);
        if (null != op) {
            Object result = op.service().evaluate(factory().emptyForm()).down();
            if (result instanceof Number) {
                return ((Number) result).intValue();
            }
            throw new AgentException(
                "Hash service does not return numerical value."
            );
        }
        return runtimeHashCode();
    }

    /**
     * Hides the specified label in this form.
    **/
    public Form hide(String label)
    {
        return this;
    }

    public final Iterator labels()
    {
        return getBindings().keySet().iterator();
    }

    /**
     * Checks whether this form is the empty form.
     *
     * @return <code>true</code> if this form is the empty form,
     *         <code>false</code> otherwise.
    **/
    public boolean isEmpty()
    {
    	return false;
    }

    /**
     * Checks whether this form is a peer form.
     *
     * @return <code>true</code> if this form is a peer form,
     *         <code>false</code> otherwise.
    **/
    public final boolean isPeer()
    {
        return unwrap() != this;
    }

    /**
     * Checks whether this form is executable (a service).
     *
     * @return <code>true</code> if this form can be executed,
     *         <code>false</code> otherwise.
    **/
    public boolean isService()
    {
        return false;
    }

    /**
     * Projects this form on the specified label.
     *
     * @param  label the label to be projected.
     *
     * @return The form bound to label in this form, i.e. returns
     *         <i>this.label</i>.
     * @throws FormProjectException if the label is not bound in this form.
    **/
    public final Form project(String label)
    {
        Form result = projectDef(label, null);
        if (null == result) {
            throw new FormProjectException(this, label);
        }
        return result;
    }

    public abstract boolean runtimeEquals(Form form);

    public abstract int runtimeHashCode();

    public abstract String runtimeToString();

    /**
     * Returns a string representation of this form.
     *
     * @return a string representation of this form.
    **/
    public String toString()
    {
        Form op = projectDef(LABEL_TOSTRING, null);
        if (null != op) {
            Object result = op.service().evaluate(factory().emptyForm().bind(LABEL_ARG)).down();
            if (result instanceof String) {
                return (String) result;
            }
            throw new AgentException(
                "String operator does not return string value."
            );
        }
        return runtimeToString();
    }

    /**
     * Returns the contractual type of this form.
    **/
    public ContractualType type()
    {
        throw new AgentException("Cannot infer type for this kind of form (" +
            getClass() + ").");
    }

    public Object unwrap()
    {
        return this;
    }

    // ------------------------------------------------------------------------
    // default
    // ------------------------------------------------------------------------

    public void push(VM vm)
    {
        vm.doPushForm(this);
    }

    // ------------------------------------------------------------------------
    // protected
    // ------------------------------------------------------------------------

    protected DefaultForm(FormFactory factory)
    {
        m_factory = factory;
    }

    protected SortedMap buildBindings()
    {
        return new TreeMap();
    }

    protected FormFactory factory()
    {
        return m_factory;
    }

    protected Object getPeer()
    {
        return null;
    }

    public Form projectDef(String label, Form defaultForm)
    {
        return defaultForm;
    }

    // ------------------------------------------------------------------------
    // private
    // ------------------------------------------------------------------------

    private SortedMap m_bindings;

    private final FormFactory m_factory;

}
